# By AWARE on RC: python3.6 and it happy this *local* version of dataclasses, but ver>3.6 should go with system-wide version
try:
    from dataclasses import dataclass
except:
    from dataclasses_local import dataclass
from typing      import Any
import functools
import time
import sys, re
import warnings

class ColorsError(Exception):
    pass

@dataclass
class Colors:
    internal_name : str = "[Colors]"

    END      = '\33[0m'

    BOLD     = '\33[1m'; ITALIC   = '\33[3m';  UNDERLINE = '\33[4m'
    BLINK    = '\33[5m'; BLINK2   = '\33[6m';  SELECTED  = '\33[7m'

    # Text
    KT = BLACK  = '\33[30m';  RT = RED      = '\33[31m'; GT = GREEN    = '\33[32m'
    YT = YELLOW = '\33[33m';  BT = BLUE     = '\33[34m'; VT = VIOLET   = '\33[35m'
    CT = CYAN   = '\33[36m';  WT = WHITE    = '\33[37m'
    # Text
    ET  = GREY    = '\33[90m'; R2T = RED2    = '\33[91m'; G2T = GREEN2  = '\33[92m'
    Y2T = YELLOW2 = '\33[93m'; B2T = BLUE2   = '\33[94m'; V2T = VIOLET2 = '\33[95m'
    C2T = CYAN2   = '\33[96m'; W2T = WHITE2  = '\33[97m'

    # Background
    TG = TRANSPARENT = ''
    KG = BLACKBG  = '\33[40m'; RG = REDBG    = '\33[41m'; GG = GREENBG  = '\33[42m'
    YG = YELLOWBG = '\33[43m'; BG = BLUEBG   = '\33[44m'; VG = VIOLETBG = '\33[45m'
    CG = CYANBG  = '\33[46m';  WG = WHITEBG  = '\33[47m'
    # Background
    EG  = GREYBG    = '\33[100m'; R2G = REDBG2    = '\33[101m'; G2G = GREENBG2  = '\33[102m'
    Y2G = YELLOWBG2 = '\33[103m'; B2G = BLUEBG2   = '\33[104m'; V2G = VIOLETBG2 = '\33[105m'
    C2G = CYANBG2   = '\33[106m'; W2G = WHITEBG2  = '\33[107m'

    WARNING = RED2
    NOTE    = CYAN2
    ERROR   = RED

    TEXT_STYLES = [
        BLACK, RED, GREEN, YELLOW, BLUE, VIOLET, CYAN, WHITE,
        GREY, RED2, GREEN2, YELLOW2, BLUE2, VIOLET2, CYAN2, WHITE2,
    ]
    BG_STYLES = [
        TRANSPARENT,
        BLACKBG, REDBG, GREENBG, YELLOWBG, BLUEBG, VIOLETBG, CYANBG, WHITEBG,
        GREYBG, REDBG2, GREENBG2, YELLOWBG2, BLUEBG2, VIOLETBG2, CYANBG2, WHITEBG2,
    ]

    POSITIVE_MEANING = ["on", "true", "yes" ,"ok", "good", "nice"]
    NEGATIVE_MEANING = ["off", "false", "no", "bad", "awful"]

    @staticmethod
    def namestr(variable, smallest=False):
        """Return all matches"""
        matches = [key for key in vars(Colors) if vars(Colors)[key] == variable]
        if not smallest: return matches
        else:
            output = matches[0]
            for match in matches:
                if len(match) < len(output): output = match
            return output

    @staticmethod
    def demo(message:str ="ABC") -> None:
        space = len(message)
        if space < 3:
            if space == 2: message = f"{message} ";
            if space == 1: message = f" {message} "
            space = len(message)
        _zr, _oc, _or = False, False, False # zero_row, once_col, once_row
        for text_style in Colors.TEXT_STYLES:  # row
            if not _oc:
                for bg_style in Colors.BG_STYLES:
                    if not _zr: print(f"{' '.join(['' for i in range(space+1)])}", end=" "); _zr = True
                    var_name = f"{Colors.namestr(bg_style,True)}"
                    while len(var_name) < space: var_name += " "
                    print(var_name, end=" ");
                    _oc = True
            print()
            for bg_style in Colors.BG_STYLES:  # col
                if not _or:
                    var_name=f"{Colors.namestr(text_style,True)}"
                    while len(var_name) < space: var_name += " "
                    print(var_name,  end=" ")
                    _or = True
                print(f"{bg_style}{text_style}{message}{Colors.END}", end= " ")
            _or = False

    @staticmethod
    def _match(word:str) -> str:
        _word = word.lower()
        for pattern_both in Colors.POSITIVE_MEANING + Colors.NEGATIVE_MEANING:
            if re.search(pattern_both, _word):
                # MATCH: positive or negative?
                for pattern_pos in Colors.POSITIVE_MEANING:
                    # positive
                    if re.search(pattern_pos, _word): return f"{Colors.GREEN2}{word}{Colors.END}"
                for pattern_neg in Colors.NEGATIVE_MEANING:
                    # negative
                    if re.search(pattern_neg, _word): return f"{Colors.RED2}{word}{Colors.END}"
        return f"{word}"

    @staticmethod
    def colourise(something:Any) -> str:
        if isinstance(something, (list, dict, tuple)):  raise ColorsError(f"Colors class does not support {type(something)} type yet.")
        # rest: float, int, str, sentence
        string = str(something)
        output = ""
        # Take this nice shoes. Nice! Thank you.
        if len(string.split()) > 1:
            for item in something.split():
                output += Colors._match(str(item)) + " "
            return output[:-1] # remove last space
        else:
            return Colors._match(string)



# Please use template
"""
def decorator(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        # Do something before
        value = func(*args, **kwargs)
        # Do something after
        return value
    return wrapper_decorator
def repeat(num_times):
    def decorator_repeat(func):
        @functools.wraps(func)
        def wrapper_repeat(*args, **kwargs):
            for _ in range(num_times):
                value = func(*args, **kwargs)
            return value
        return wrapper_repeat
    return decorator_repeat
"""

@dataclass
class InternalError:
    internal_name   : str = "[IE]"
    _internal_error : int = 0
    def get(self):            return self._internal_error
    def set(self, value:int): self._internal_error = value
    def occurred(self):       self._internal_error += 1; print(f"{Colors.RED}{self.internal_name} Occurred!{Colors.END}")
    def reset(self):          self._internal_error = 0
    def is_occurred(self):    return self._internal_error > 0


def with_no_internal_error(_func=None, *, message:str = ""):
    """
        Be sure that no internal errors occurred before execution of the function
        Requested:
            * InternalError class
        Optional:
            * message for printing
        """
    def decorator_with_no_internal_error(func):
        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):
            try:
                instance = args[0].internal_error
            except Exception as e:
                print(f"Problem: {e}. Did you imported InternalError class?")
                sys.exit(1)
            internal_error = instance.get()
            if not internal_error:
                if message: print(message)
                return func(*args, **kwargs)
            else:
                print(f"{Colors.RED}FATAL! Until the call of `{func.__name__}` function. Internal error occurred already [{internal_error}] times. Read the message above!{Colors.END}")
        return wrapper_decorator
    if _func is None:  return decorator_with_no_internal_error
    else:              return decorator_with_no_internal_error(_func)


def slow_down(func):
    """Sleep 1 second before calling the function"""
    @functools.wraps(func)
    def wrapper_slow_down(*args, **kwargs):
        time.sleep(1)
        return func(*args, **kwargs)
    return wrapper_slow_down


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)
    return new_func
